//
// Porting and modifications made by Rayzzor Studios, 2011
//

/**
 * Copyright (c) 2005-2009 voidware ltd.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to
 * deal in the Software without restriction, including without limitation the
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 * sell copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS," WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 * IN THE SOFTWARE.
 */

/**
 * Here is the BCD float class.
 *
 * the BCD float is made of `P' 4dec digits. a 4dec digit is a chunk
 * of 4 decimal digits stored a a machine `short'. rather than use
 * a nibble per digit and base 10, we are using base 10,000 and thus
 * work with 4 decimal digits in one go.
 *
 * this is much more efficient because, not only is it faster, but
 * we are able to leverage the 16*16->32 and 32/16->16 multiply and
 * divide integer operations found on most modern processors as well
 * as 16 bit add & subtract.
 *
 * the exponent is stored in the last slot (ie P+1) and ranges from 
 * -10000 to +9999. the mantissa is stored unsigned and the sign is the
 * top bit of the exponent.
 *
 * there are a number of important considerations for this representation:
 *
 * the postion of the radix must always correspond with the end of a 4dec
 * word. therefore up to one extra 4dec is wasted in the representation.
 * for example, the number 12345.67 will store as [0001][2345].[6700] and
 * utilise 3 4decs rather than two. this is the drawback of using 4decs.
 * bytes would waste up to 1 nibble and nibbles dont waste anything which
 * is why they are normally chosen for short digit (eg 6) representations.
 * for long digit (eg 20) this method is preferable.
 *
 * whence we take P=6 which always has the capacity for 21 decimal digits
 * of mantissa (and up to 24 when it gets lucky) and at P=7 we have
 * 1 spare 4dec word.
 *
 * the trick of storing the mantissa as the P+1'th slot is cunning.
 * most of the arithmetic proceeds by extracting the sign and exponent
 * before working with the mantissa. thus for the purposes of the internal
 * calculations, we can let the mantissa spill out into the last `digit'
 * and effectively have one extra digit of working precision. this is
 * especially useful for making sure we get the correct rounding at the
 * end when we put back the exponent. and also for accommodating any overall
 * numeric carry overflow before the number is shifted down.
 *
 * it has been suggested that, rather than use an extra 16 bit word, 5 words
 * are used for 20 digits without waste. this is possible but it involves
 * multiplying all 5 terms by 10, 100 or 1000 with carry, in order to align
 * the radix for operation. i didnt want this overhead, but it is true that
 * it would result in smaller representation. but dont forget that the
 * overall representation in size will be rounded up to a multiple of 4 bytes
 * anyway.
 *
 * lastly, floating point code is notorious for hiding bugs, often for years.
 * this code is new and there will be some "simple" cases that come out
 * plain wrong. thats the way with floating point. even CPU makers get it
 * wrong sometimes.
 */


using System;


namespace WP7_42s
{
    public static class GlobalMembersBcdfloat
    {
        public static int[] BCDDecade = { 1000, 100, 10, 1 };


        // #define ISEXP(_c) ((_c) == 'e' || (_c) == 'E')
        public static bool ISEXP(char c) { return (c == 'e' || c == 'E'); }

        // #define GET_EXP(_d, _p) (((short)((_d)[_p] << 3)) >> 3)
        public static int GET_EXP(ushort[] d, int p) { return (((short)((d)[p] << 3)) >> 3); }

        // #define SET_EXP(_d, _p, _v) ((_d)[_p] = (_v) & EXPMASK)
        public static void SET_EXP(ushort[] d, int p, int v) { d[p] = (ushort)(v & App.EXPMASK); }

        // #define CLEAR_SIGN(_d, _p) ((_d)[_p] &= ~NEG)
        public static void CLEAR_SIGN(ushort[] d, int p)
        {
            ushort n = App.NEG;
            d[p] &= (ushort)~n;
        }

        // #define NEGATE_SIGN(_d, _p) ((_d)[_p] ^= NEG)
        public static void NEGATE_SIGN(ushort[] d, int p) { d[p] ^= App.NEG; }

        // #define GET_SPECIAL(_d, _p) ((_d)[_p]&0x6000)
        public static bool GET_SPECIAL(ushort[] d, int p) { return (d[p] & 0x6000) != 0; }

        // #define GET_NAN(_d, _p) ((_d)[_p]&NAN_EXP)
        public static bool GET_NAN(ushort[] d, int p) { return (d[p] & App.NAN_EXP) != 0; }

        // #define GET_INF(_d, _p) ((_d)[_p]&0x2000)
        public static bool GET_INF(ushort[] d, int p) { return (d[p] & 0x2000) != 0; }

        // #define GET_NEG_BIT(_d, _p) ((_d)[_p]&NEG)
        public static bool GET_NEG_BIT(ushort[] d, int p) { return (d[p] & App.NEG) != 0; }

        // zero assuming normal (non-special)
        // #define GET_ZERO_NORM(_d, _p) ((_d)[0] == 0)
        public static bool GET_ZERO_NORM(ushort[] d, int p) { return d[0] == 0; }

        // is zero generally
        // #define IS_ZERO(_d, _p) (GET_ZERO_NORM(_d, _p) && GET_SPECIAL(_d, _p) == 0)
        public static bool IS_ZERO(ushort[] d, int p) { return GET_ZERO_NORM(d, p) && !GET_SPECIAL(d, p); }

        // negative assuming non-special
        // #define GET_NEG_NORM(_d, _p) (GET_NEG_BIT(_d, _p) && !GET_ZERO_NORM(_d,_p))
        public static bool GET_NEG_NORM(ushort[] d, int p) { return GET_NEG_BIT(d, p) && !GET_ZERO_NORM(d, p); }




        public static int bcd_round25(ushort[] d, int pn)
        {
            // round d_[P] into the mantissa and mask off digits after 25.
            int i = pn - 1;
            int v;

            if (d[0] < 10)
                v = d[i] + ((d[pn] >= 5000) ? 1 : 0);
            else if (d[0] < 100)
                v = ((((int)d[i] + 5) * 3277) >> 15) * 10;
            else if (d[0] < 1000)
                v = ((((int)d[i] + 50) * 5243) >> 19) * 100;
            else
                v = ((((int)d[i] + 500) * 8389) >> 23) * 1000;

            while (v >= App.BASE)
            {
                d[i] = (ushort)(v - App.BASE);
                if (i == 0)
                {
                    // shift
                    for (i = pn; i > 0; --i)
                        d[i] = d[i - 1];
                    d[0] = 1;
                    return 1;
                }
                v = d[--i] + 1;
            }
            d[i] = (ushort)v;
            return 0;
        }

        public static int bcd_round(ushort[] d, int pn)
        {
            // round d_[P] into the mantissa
            if (d[pn] >= 5000)
            {
                int i = pn - 1;
                int v = d[i] + 1;
                while (v >= App.BASE)
                {
                    d[i] = (ushort)(v - App.BASE);
                    if (i == 0)
                    {
                        // shift
                        for (i = pn; i > 0; --i)
                            d[i] = d[i - 1];
                        d[0] = 1;
                        return 1;
                    }
                    v = d[--i] + 1;
                }
                d[i] = (ushort)v;
            }
            return 0;
        }



        public static void bcd_uadd(ushort[] a, ushort[] b, ushort[] c, int pn)
        {
            int ea = GET_EXP(a, pn);
            int eb = GET_EXP(b, pn);
            int d = ea - eb;
            int i;

            if (d <= pn) // otherwise `b' is insignificant
            {
                int ca;
                int v;
                int j = pn - d;
                i = pn - 1;

                // copy in first insignificant digit used by final rounding
                v = 0;
                if (d > 0)
                    v = b[j];
                c[pn] = (ushort)v;

                ca = 0;
                while (j > 0) // perform addition of overlapping terms
                {
                    v = a[i] + b[--j] + ca;
                    ca = 0;
                    if (v >= App.BASE)
                    {
                        v -= App.BASE;
                        ca = 1;
                    }
                    c[i] = (ushort)v;
                    --i;
                }

                while (i >= 0) // remainder non-overlap terms
                {
                    v = a[i] + ca;
                    ca = 0;
                    if (v >= App.BASE)
                    {
                        v -= App.BASE;
                        ca = 1;
                    }
                    c[i] = (ushort)v;
                    --i;
                }

                if (ca != 0)
                {
                    // overall carry, shift down and round
                    for (i = pn; i > 0; --i)
                        c[i] = c[i - 1];
                    c[0] = (ushort)ca;
                    ++ea;
                }

                if (bcd_round25(c, pn) != 0)
                    ++ea;

                if (ea > App.BASE / 4)
                {
                    // overflow
                    for (i = 0; i < pn; ++i)
                        c[i] = 0;
                    ea = App.POS_INF_EXP;
                }

                SET_EXP(c, pn, ea);
            }
            else
            {
                for (i = 0; i <= pn; ++i)
                    c[i] = a[i];

                CLEAR_SIGN(c, pn);
            }
        }

        public static void bcd_usub(ushort[] a, ushort[] b, ushort[] c, int pn)
        {
            int ea = GET_EXP(a, pn);
            int eb = GET_EXP(b, pn);
            bool neg = false;
            int i;

            int d = ea - eb;
            if (d <= pn)
            {
                int ca;
                int v;

                int j = pn - d;
                i = pn - 1;

                ca = 0;
                v = 0;

                // first insignificant digit my be used for rounding
                if (d > 0)
                {
                    v = b[j];
                    if (v != 0)
                    {
                        v = App.BASE - v;
                        ca = 1;
                    }
                }
                c[pn] = (ushort)v;

                while (j > 0)
                {
                    v = a[i] - b[--j] - ca;
                    ca = 0;
                    if (v < 0)
                    {
                        ca = 1;
                        v += App.BASE;
                    }
                    c[i] = (ushort)v;
                    --i;
                }

                if (ca != 0)
                {
                    while (i >= 0)
                    {
                        v = a[i];
                        if (v != 0)
                        {
                            // carry absorbed
                            c[i] = (ushort)(v - 1);

                            // copy remainder
                            while (i > 0)
                            {
                                --i;
                                c[i] = a[i];
                            }

                            // NB: i == 0
                            break;
                        }
                        else
                            c[i--] = App.BASE - 1;
                    }

                    if (i < 0) // carry all the way up
                    {
                        // overall borrow, need to complement number
                        for (i = pn; i >= 0; --i)
                        {
                            v = App.BASE - 1 - c[i] + ca;
                            ca = 0;
                            if (v >= App.BASE)
                            {
                                ca = 1;
                                v -= App.BASE;
                            }
                            c[i] = (ushort)v;
                        }
                        neg = true;
                    }
                }
                else
                {
                    // copy remainder of number over
                    while (i >= 0)
                    {
                        c[i] = a[i];
                        --i;
                    }
                }

                int e = ea;
                i = 0;
                while (c[i > pn ? pn : i] == 0 && i <= pn)      // TO DO: added bounds check (i>pn ? pn:i)
                    i++;                                        // Since we're checking for zeros and c[pn] was already tested, this should be ok - verify

                if (i > 0)
                {
                    if (i == pn + 1)
                    {
                        // is zero
                        e = 0;
                    }
                    else
                    {
                        e -= i;
                        if (e <= -App.BASE / 4)
                        {
                            // underflow
                            c[0] = 0;
                            e = 0;
                        }
                        else
                        {
                            for (j = 0; j <= pn - i; j++)
                                c[j] = c[j + i];
                            for (; j <= pn; j++)
                                c[j] = 0;
                        }
                    }
                }

                if (bcd_round25(c, pn) != 0)
                    ++e;

                if (e > App.BASE / 4) // not sure this can happen
                {
                    for (i = 0; i < pn; ++i)
                        c[i] = 0;
                    ea = App.POS_INF_EXP;
                }
                SET_EXP(c, pn, e);
                if (neg)
                    NEGATE_SIGN(c, pn);
            }
            else
            {
                // `b' is insignificant
                for (i = 0; i <= pn; ++i)
                    c[i] = a[i];
                CLEAR_SIGN(c, pn);
            }
        }

        public static void bcd_add(ushort[] a, ushort[] b, ushort[] c, int pn)
        {
            int i;
            if (GET_SPECIAL(a, pn) | GET_SPECIAL(b, pn))
            {
                for (i = 0; i < pn; ++i)
                    c[i] = 0;

                c[pn] = App.NAN_EXP;

                //         inf + inf = inf
                //         -inf + (-inf) = -inf
                //
                if (!(GET_NAN(a, pn) | GET_NAN(b, pn)))
                {
                    if (GET_INF(a, pn))
                    {
                        if (!GET_INF(b, pn) || (a[pn] == b[pn]))
                        {
                            // inf + x = inf
                            c[pn] = a[pn];
                        }
                    }
                    else
                    {
                        // b is inf
                        c[pn] = b[pn];
                    }
                }
            }
            else
            {
                if (GET_ZERO_NORM(a, pn))
                {
                    for (i = 0; i <= pn; ++i)
                        c[i] = b[i];
                    return;
                }
                if (GET_ZERO_NORM(b, pn))
                {
                    for (i = 0; i <= pn; ++i)
                        c[i] = a[i];
                    return;
                }

                int ea = GET_EXP(a, pn);
                int eb = GET_EXP(b, pn);

                bool na = GET_NEG_NORM(a, pn);
                bool nb = GET_NEG_NORM(b, pn);

                if (na != nb) // sub
                {
                    if (ea >= eb)
                        bcd_usub(a, b, c, pn);
                    else
                    {
                        bcd_usub(b, a, c, pn);
                        na = nb;
                    }
                }
                else
                {
                    if (ea >= eb)
                        bcd_uadd(a, b, c, pn);
                    else
                        bcd_uadd(b, a, c, pn);
                }
                if (na)
                    NEGATE_SIGN(c, pn);
            }
        }

        public static void bcd_sub(ushort[] a, ushort[] b, ushort[] c, int pn)
        {
            int i;
            if (GET_SPECIAL(a, pn) | GET_SPECIAL(b, pn))
            {
                for (i = 0; i < pn; ++i)
                    c[i] = 0;

                // all others -> nan
                c[pn] = App.NAN_EXP;

                if (!(GET_NAN(a, pn) | GET_NAN(b, pn)))
                {
                    if (GET_INF(a, pn))
                    {
                        if (!GET_INF(b, pn) || (a[pn] != b[pn]))
                        {
                            //                     -inf - (inf) = -inf
                            //                     inf - (-inf) = inf
                            //
                            c[pn] = a[pn];
                        }
                    }
                    else
                    {
                        // b is inf
                        c[pn] = App.NEG_INF_EXP;
                    }
                }
            }
            else
            {
                if (GET_ZERO_NORM(a, pn))
                {
                    for (i = 0; i <= pn; ++i)
                        c[i] = b[i];
                    NEGATE_SIGN(c, pn);
                    return;
                }

                if (GET_ZERO_NORM(b, pn))
                {
                    for (i = 0; i <= pn; ++i)
                        c[i] = a[i];
                    return;
                }

                bool na = GET_NEG_NORM(a, pn);
                bool nb = GET_NEG_NORM(b, pn);

                int ea = GET_EXP(a, pn);
                int eb = GET_EXP(b, pn);

                if (na == nb) // sub
                {
                    if (ea >= eb)
                    {
                        bcd_usub(a, b, c, pn);
                        if (na)
                            NEGATE_SIGN(c, pn);
                    }
                    else
                    {
                        bcd_usub(b, a, c, pn);
                        if (!na)
                            NEGATE_SIGN(c, pn);
                    }
                }
                else
                {
                    if (ea >= eb)
                        bcd_uadd(a, b, c, pn);
                    else
                        bcd_uadd(b, a, c, pn);
                    if (na)
                        NEGATE_SIGN(c, pn);
                }
            }
        }

        public static void bcd_mul(ushort[] a, ushort[] b, ushort[] c, int pn)
        {
            // zero flags assuming not special
            bool az = GET_ZERO_NORM(a, pn);
            bool bz = GET_ZERO_NORM(b, pn);

            // neg flag assuming non-zero and non-special
            bool na = GET_NEG_BIT(a, pn);
            bool nb = GET_NEG_BIT(b, pn);

            int i;
            for (i = 0; i <= pn; ++i)
                c[i] = 0;

            if (GET_SPECIAL(a, pn) | GET_SPECIAL(b, pn))
            {
                // all others -> nan
                c[pn] = App.NAN_EXP;

                if (!(GET_NAN(a, pn) | GET_NAN(b, pn)))
                {
                    if ((GET_INF(a, pn) && (GET_INF(b, pn) || (!bz))) || !az)
                    {
                        // inf * inf -> inf
                        // inf * non-zero -> inf
                        // non-zero * inf -> inf
                        c[pn] = App.POS_INF_EXP;
                        if (na != nb)
                            NEGATE_SIGN(c, pn);
                    }
                }
            }
            else if (!az && !bz)
            {
                int ca;
                int j;
                int u;
                int v;

                int ea = GET_EXP(a, pn);
                int eb = GET_EXP(b, pn);

                ushort[] acc = new ushort[2 * (App.P + 1) + 1];
                for (i = 0; i <= pn; ++i)
                    acc[i] = 0;

                for (i = pn - 1; i >= 0; --i)
                {
                    bcd_round(c, pn);
                    for (j = pn; j > 0; --j)
                        c[j] = c[j - 1];
                    c[0] = 0;

                    u = a[i];
                    if (u == 0)
                        continue;

                    ca = 0;
                    for (j = pn; j > 0; --j)
                    {
                        v = b[j - 1] * u + ca;
                        ca = 0;
                        if (v >= App.BASE)
                        {
                            ca = v / App.BASE;
                            v = v - ca * App.BASE;
                        }
                        acc[j] = (ushort)v;
                    }
                    acc[0] = (ushort)ca;

                    // now add acc into c
                    ca = 0;
                    for (j = pn; j >= 0; --j)
                    {
                        v = c[j] + acc[j] + ca;
                        ca = 0;
                        if (v >= App.BASE)
                        {
                            ca = 1;
                            v -= App.BASE;
                        }
                        c[j] = (ushort)v;
                    }
                    // won't be any overall carry
                }

                if (c[0] == 0)
                {
                    for (i = 0; i < pn; ++i)
                        c[i] = c[i + 1];
                    c[pn] = 0;
                }
                else
                {
                    ++ea;
                    if (bcd_round25(c, pn) != 0)
                        ++ea;
                }

                ea += eb - 1;
                if (ea <= -App.BASE / 4 || ea > App.BASE / 4)
                {
                    for (i = 0; i <= pn; ++i) // underflow
                        c[i] = 0;
                    if (ea > App.BASE / 4) // overflow
                        c[pn] = App.POS_INF_EXP;
                }
                else
                {
                    SET_EXP(c, pn, ea);

                    // fix sign
                    if (na != nb)
                        NEGATE_SIGN(c, pn);
                }
            }
        }

        public static void bcd_div(ushort[] a, ushort[] b, ushort[] c, int pn)
        {
            bool @as = GET_SPECIAL(a, pn);
            bool bs = GET_SPECIAL(b, pn);

            // zero flags assuming not special
            bool az = GET_ZERO_NORM(a, pn) && !@as;
            bool bz = GET_ZERO_NORM(b, pn) && !bs;

            // neg assuming non-special
            bool na = GET_NEG_BIT(a, pn) && !az;
            bool nb = GET_NEG_BIT(b, pn) && !bz;
            int i;

            if (@as | bs | az | bz)
            {
                for (i = 0; i < pn; ++i)
                    c[i] = 0;

                // all others -> nan
                c[pn] = App.NAN_EXP;

                if (!(GET_NAN(a, pn) | GET_NAN(b, pn)))
                {
                    if (GET_INF(a, pn) | bz)
                    {
                        // inf/inf -> nan
                        if (!GET_INF(b, pn) && !az)
                        {
                            // inf/x -> inf
                            if (na == nb)
                                c[pn] = App.POS_INF_EXP;
                            else
                                c[pn] = App.NEG_INF_EXP;
                        }
                    }
                    else if (GET_INF(b, pn) | az)
                    {
                        // x/inf -> 0
                        c[pn] = 0;
                    }
                }
            }
            else
            {
                int u;
                int v;
                int ca;
                int j = 0;
                int q;

                int ea = GET_EXP(a, pn);
                int eb = GET_EXP(b, pn);

                ushort[] acc = new ushort[2 * (App.P + 1) + 1];
                ushort[] b1 = new ushort[2 * (App.P + 1) + 1];

                u = App.BASE / (b[0] + 1);

                if (u != 1)
                {
                    // prenormialise `a' and move into acc using spare digit
                    ca = 0;
                    for (i = pn; i > 0; --i)
                    {
                        v = a[i - 1] * u + ca;
                        ca = 0;
                        if (v >= App.BASE)
                        {
                            ca = v / App.BASE;
                            v -= ca * App.BASE;
                        }
                        acc[i] = (ushort)v;
                    }
                    acc[0] = (ushort)ca;

                    // prenormalise `b' into b1
                    ca = 0;
                    for (i = pn - 1; i >= 0; --i)
                    {
                        v = b[i] * u + ca;
                        ca = 0;
                        if (v >= App.BASE)
                        {
                            ca = v / App.BASE;
                            v -= ca * App.BASE;
                        }
                        b1[i] = (ushort)v;
                    }
                }
                else
                {
                    // u is often 1
                    for (i = pn - 1; i >= 0; --i)
                    {
                        acc[i + 1] = a[i];
                        b1[i] = b[i];
                    }
                    acc[0] = 0;
                }

                for (; ; )
                {
                    if (acc[0] == b1[0])
                        q = App.BASE - 1;
                    else
                    {
                        v = acc[0] * ((int)App.BASE) + acc[1];
                        q = v / b1[0];
                        while (b1[1] * q > ((v - q * b1[0]) * App.BASE + acc[2]))
                            --q;
                    }

                    if (q == 0 && j == 0)
                    {
                        //                 first quotient digit is zero. can gain extra
                        //                 accuracy by ignoring this and adjusting exponent.
                        //
                        --ea;
                    }
                    else
                    {
                        if (j == pn)
                        {
                            c[j] = (ushort) q;
                            break;
                        }

                        ca = 0;
                        for (i = pn; i > 0; --i)
                        {
                            v = acc[i] - b1[i - 1] * q - ca;
                            ca = 0;
                            if (v < 0)
                            {
                                ca = (-v + App.BASE - 1) / App.BASE;
                                v += ca * App.BASE;
                            }
                            acc[i] = (ushort)v;
                        }
                        v = acc[0] - ca;

                        if (v != 0)
                        {
                            // the infamous add back correction
                            ca = 0;
                            for (i = pn; i > 0; --i)
                            {
                                v = acc[i] + b1[i - 1] + ca;
                                ca = 0;
                                if (v >= App.BASE)
                                {
                                    ca = 1;
                                    v -= App.BASE;
                                }
                                acc[i] = (ushort)v;
                            }
                            q--;
                        }
                        if (q == 0 && j == 0)
                            --ea;
                        else
                            c[j++] = (ushort)q;
                    }

                    // left shift
                    for (i = 0; i < pn; ++i)
                        acc[i] = acc[i + 1];
                    acc[pn] = 0;
                }

                if (bcd_round25(c, pn) != 0)
                    ++ea;

                ea -= eb - 1;
                if (ea <= -App.BASE / 4 || ea > App.BASE / 4)
                {
                    for (i = 0; i <= pn; ++i)
                        c[i] = 0;
                    if (ea > App.BASE / 4)
                        c[pn] = App.POS_INF_EXP;
                }
                else
                    SET_EXP(c, pn, ea);
                if (na != nb)
                    NEGATE_SIGN(c, pn);
            }
        }



        public static int bcd_cmp(ushort[] a, ushort[] b, int pn)
        {
            if (GET_NAN(a, pn) || GET_NAN(b, pn))
            {
                // NaNs are not equal to anything, even themselves;
                // ALL comparisions involving them should return 'false'
                // (except '!=', which should return 'true').
                // So, I return a special value here.
                return 2;
            }

            // zero flags assuming not special
            bool az = GET_ZERO_NORM(a, pn) && !GET_INF(a, pn);
            bool bz = GET_ZERO_NORM(b, pn) && !GET_INF(b, pn);

            // neg
            bool na = GET_NEG_BIT(a, pn) && !az;
            bool nb = GET_NEG_BIT(b, pn) && !bz;

            if (na != nb)
                return na ? -1 : 1;

            // If a and b are both negative, switch them. This way, I can
            // ignore signs in the remainder of this method.
            if (na)
            {
                ushort[] t = a;
                a = b;
                b = t;
            }

            if (GET_INF(a, pn))
                return GET_INF(b, pn) ? 0 : 1;

            if (GET_INF(b, pn))
                return -1;

            if (az)
                return bz ? 0 : -1;

            if (bz)
                return 1;

            int ea = GET_EXP(a, pn);
            int eb = GET_EXP(b, pn);
            if (ea != eb)
                return ea < eb ? -1 : 1;

            for (int i = 0; i < pn; i++)
            {
                int da = a[i];
                int db = b[i];
                if (da != db)
                    return da < db ? -1 : 1;
            }
            return 0;
        }

        public static void bcd_fromUInt(ushort[] d, int pn, uint v)
        {
            // quicker to deal with cases separately
            if (v < App.BASE)
            {
                d[0] = (ushort)v;
                d[pn] = 1;
            }
            else if (v < ((int)App.BASE) * App.BASE)
            {
                d[0] = (ushort)(v / App.BASE);
                d[1] = (ushort)(v - d[0] * App.BASE);
                d[pn] = 2;
            }
            else
            {
                d[0] = (ushort)(v / (((int)App.BASE) * App.BASE));
                v -= (uint)(d[0] * (((int)App.BASE) * App.BASE));
                d[1] = (ushort)(v / App.BASE);
                d[2] = (ushort)(v - d[1] * App.BASE);
                d[pn] = 3;
            }
        }

        internal static uint isqrt(uint v)
        {
            /* integer root for x<= 9999 */
            uint x = 0;
            int b = 1 << 7;
            do
            {
                x ^= (uint)b;
                if (x * x > v)
                    x ^= (uint)b;
            } while ((b >>= 1) != 0);
            return x;
        }
    }





    public class BCDFloatData
    {
        //     store P 4dec `digits', equivalent to P*4 decimal digits.
        //     the last place is the exponent.
        //     
        public ushort[] d_ = new ushort[App.P + 1];
    }







    public class BCDFloat : BCDFloatData
    {
        public enum Format
	    {
		    format_normal = 0,
		    format_scimode = 1,
		    format_truncate = 2, // no rounding
	    }


        //
        // Constructors
        //
        #region Constructors

        public BCDFloat(int d0, int d1, int d2, int d3, int d4, int d5, int d6, int d7)
        {
            d_[0] = (ushort)d0;
            d_[1] = (ushort)d1;
            d_[2] = (ushort)d2;
            d_[3] = (ushort)d3;
            d_[4] = (ushort)d4;
            d_[5] = (ushort)d5;
            d_[6] = (ushort)d6;
            d_[7] = (ushort)d7;
        }

        public BCDFloat(ushort[] d)
        {
            if (d.Length == App.P + 1)
            {
                d_[0] = d[0];
                d_[1] = d[1];
                d_[2] = d[2];
                d_[3] = d[3];
                d_[4] = d[4];
                d_[5] = d[5];
                d_[6] = d[6];
                d_[7] = d[7];
            }
        }

        public BCDFloat() { }            // Warning: not initialised.

        public BCDFloat(string s)
        {
            s += "\x0000";

            // make a BCD float from a string.
            //
            _init();

            int _s = 0;

            bool neg = false;
            if (s[_s] == '-')
            {
                neg = true;
                ++_s;
            }

            int i;
            int j = 0;
            int e = 0;
            int p;

            // find the end of the input string
            int endp = _s;
            int startp = -1;
            int point = -1;
            bool begun = false;

            while ((s[endp] != '\x0000') && !GlobalMembersBcdfloat.ISEXP(s[endp]))
            {
                if (s[endp] == '.')
                {
                    if (point != -1)
                        break;

                    point = endp;

                    if (startp == -1)
                    {
                        startp = endp;
                        begun = true;
                    }
                }
                else
                {
                    if (s[endp] < '0' || s[endp] > '9')
                        break;

                    if (s[endp] != '0' && !begun)
                    {
                        startp = endp;
                        begun = true;
                    }
                }

                ++endp;
            }

            if (startp == -1)
            {
                // check for Inf before fail
                //if (!strnicmp(s, "inf", 3))
                if (s.Length >= 3)
                    if ((s[0] == 'i' || s[0] == 'I') && (s[1] == 'n' || s[1] == 'N') && (s[2] == 'f' || s[2] == 'F'))
                    {
                        _bcdUpdate(posInfD_);

                        if (neg)
                            negate();
                    }

                return;
            }


            bool eneg = false;

            if (startp != -1)
            {
                if (GlobalMembersBcdfloat.ISEXP(s[endp]))
                {
                    p = endp + 1;

                    if (s[p] == '-')
                    {
                        eneg = true;
                        ++p;
                    }
                    else if (s[p] == '+')
                        ++p;  // allow E+1234 

                    e = 0;


                    // Skip leading zeroes
                    char c;
                    while ((c = s[p++]) == '0') ;

                    // Accept up to 6 digits; the reason why I don't stop after
                    // 5 digits is because I want to make sure that 0.1e100000
                    // is converted to Inf, and not 1e9999
                    for (i = 0; i < 6; ++i)
                    {
                        if (c >= '0' && c <= '9')
                            e = e * 10 + c - '0';
                        else
                            break;

                        c = s[p++];
                    }

                    if (eneg)
                        e = -e;
                }

                /* represent the decimal point by adjusting the exponent */
                if (point != -1)
                    e += (int)(point - startp);
                else
                    e += (int)(endp - startp);

                /* calculate the decade offset of the exponent remainder */
                if (e >= 0)
                {
                    i = (e & 3);
                    if (i != 0)
                        i = 4 - i;
                }
                else
                    i = ((-e) & 3);

                e += i;


                /* convert to 4dec */
                e >>= 2;

                p = startp;
                bool leading_zero = true;

                while (p != endp)
                {
                    int d = 0;

                    while (i < 4)
                    {
                        if (s[p] != '.')
                        {
                            d += (s[p] - '0') * GlobalMembersBcdfloat.BCDDecade[i];
                            ++i;
                        }

                        ++p;
                        if (p == endp)
                            break;
                    }

                    i = 0;
                    if (leading_zero && d == 0)
                    {
                        e--;
                        continue;
                    }

                    leading_zero = false;
                    d_[j] = (ushort)d;

                    if (++j > App.P)
                        // full up.
                        break;
                }

                if (leading_zero)
                    e = 0;

                while (j <= App.P)
                    d_[j++] = 0;

                if (_round25() != 0)
                    e++;

                if (e <= -App.BASE / 4)
                    _init();
                else
                {
                    if (e > App.BASE / 4)
                        _bcdUpdate(posInfD_);
                    else
                        d_[App.P] = (ushort)(e & App.EXPMASK);

                    if (neg)
                        negate();
                }
            }
        }

        public BCDFloat(int v)
        {
            _init();
            if (v != 0)
            {
                bool neg = v < 0;
                if (neg)
                    v = -v;
                _fromUInt((uint)v);
                if (neg)
                    negate();
            }
        }

        public BCDFloat(Int64 n)
        {
            _init();
            int i;
            int z = 5;
            bool neg = n < 0;
            if (neg)
                n = -n;
            for (i = 4; i >= 0; i--)
            {
                int d = (int)(n % App.BASE);
                d_[i] = (ushort) d;
                n /= App.BASE;
                if (d != 0)
                    z = i;
            }
            if (z < 5)
            {
                if (z > 0)
                {
                    for (i = 0; i < 5 - z; i++)
                        d_[i] = d_[i + z];
                    for (i = 5 - z; i < 5; i++)
                        d_[i] = 0;
                }
                d_[App.P] = (ushort) (5 - z);
                if (neg)
                    negate();
            }
        }

        public BCDFloat(double d)
        {
            if (GlobalMembersMathFudge.isnan(d))
            {
                _bcdUpdate(nanD_);
                return;
            }

            if (GlobalMembersMathFudge.isinf(d) != 0)
            {
                _bcdUpdate(GlobalMembersMathFudge.isinf(d) > 0 ? posInfD_ : negInfD_);
                return;
            }

            // Return exact results if d is representable exactly as a long long.
            // TO DO: there are additional cases we should try to handle exactly, e.g.
            // 0.375. Basically, the idea would be to do the conversion by summing
            // exact equivalents for the values of d's nonzero bits, just like the
            // way the table-based binary-to-bcd conversion in the Free42 Binary
            // build works. This approach returns exact results for a very large set
            // of cases, including all representable integers.

            Int64 n = (Int64)d;

            if (d == n)
            {
                _bcdUpdate(new BCDFloat(n));
                return;
            }

            bool neg = d < 0;
            if (neg)
                d = -d;

            if (d == 0)
                _init();
            else
            {
                int exp = (((int)Math.Floor(Math.Log10(d))) & ~3) >> 2;
                d_[App.P] = (ushort)((exp + 1) & App.EXPMASK);

                double m = Math.Pow(100.0, (double)exp);
                d /= m;
                d /= m;

                for (int i = 0; i < App.P; i++)
                {
                    short s = (short)d;
                    d_[i] = (ushort)s;
                    d = (d - s) * 10000;
                }

                if (neg)
                    negate();
            }

            if (_round25() != 0)
            {
                // increase exponent
                int e = (((short)(d_[App.P] << 3)) >> 3);
                ++e;

                if (e == (App.BASE / 4))
                {
                    // Out of range; return Inf
                    _bcdUpdate(!neg ? posInfD_ : negInfD_);
                }
                else
                {
                    d_[App.P] = (ushort)(e & App.EXPMASK);
                }
            }
        }

        public BCDFloat(uint v)
        {
            _init();
            if (v != 0)
                _fromUInt(v);
        }

        public BCDFloat(BCDFloatData d)
        {
            _bcdUpdate(d);
        }

        #endregion



        public static ushort[] posInfD_ = new ushort[App.P + 1] { 0, 0, 0, 0, 0, 0, 0, App.POS_INF_EXP };
        public static ushort[] negInfD_ = new ushort[App.P + 1] { 0, 0, 0, 0, 0, 0, 0, 0xA000 };
        public static ushort[] nanD_ = new ushort[App.P + 1] { 0, 0, 0, 0, 0, 0, 0, 0x4000 };



        #region InstMembers

        
        public void _bcdUpdate(ushort[] d)
        {
            if (d.Length == App.P + 1)
            {
                d_[0] = d[0];
                d_[1] = d[1];
                d_[2] = d[2];
                d_[3] = d[3];
                d_[4] = d[4];
                d_[5] = d[5];
                d_[6] = d[6];
                d_[7] = d[7];
            }
        }

        public void _bcdUpdate(BCDFloat d)
        {
            d_[0] = d.d_[0];
            d_[1] = d.d_[1];
            d_[2] = d.d_[2];
            d_[3] = d.d_[3];
            d_[4] = d.d_[4];
            d_[5] = d.d_[5];
            d_[6] = d.d_[6];
            d_[7] = d.d_[7];
        }

        public void _bcdUpdate(BCDFloatData d)
        {
            d_[0] = d.d_[0];
            d_[1] = d.d_[1];
            d_[2] = d.d_[2];
            d_[3] = d.d_[3];
            d_[4] = d.d_[4];
            d_[5] = d.d_[5];
            d_[6] = d.d_[6];
            d_[7] = d.d_[7];
        }



        public void asString(ref string buf)
        {
            _asString(ref buf, Format.format_normal, 0);
        }

        public void asStringFmt(ref string buf, Format fmt, int precision)
        {
            _asString(ref buf, fmt, precision);
        }

        public void _asString(ref string buf, Format fmt, int precision)
        {
            string p = "";

            if (isSpecial())
            {
                if (isNan())
                    p += "NaN";
                else
                { // inf
                    if (neg())
                        p += "-";

                    p += "Inf";
                }
            }
            else
            {
                int i;

                bool scimode = (fmt & Format.format_scimode) != 0;
                if (neg())
                    p += "-";

                int point = 0;
                int v;
                int eadj = 0;
                int e = exp();
                int pr = 0;  // significant digits printed
                i = 0;

                if (e > 0)
                {
                    /* if the exponent is less than our digits, we can
                     * print out the number like an integer.
                     */
                    if (e <= 5 && !scimode)
                    {
                        while (i < e)
                        {
                            v = d_[i];

                            int ptmp = p.Length;

                            if (i == 0)
                                p += v.ToString("D");
                            else
                                p += v.ToString("D4");

                            pr += p.Length - ptmp;

                            ++i;
                        }
                        e -= i;
                    }
                    else
                    {
                        /* otherwise we have a larger number. print out
                         * as scientific form.
                         */
                        v = d_[0];
                        string tb = "";

                        tb += v.ToString("D");

                        int ptmp = p.Length;
                        p += tb.Insert(1, ".");

                        ++pr;
                        point = ptmp + 1;

                        pr += tb.Length - 1;
                        eadj += tb.Length - 1;

                        ++i;
                        --e;
                    }
                }
                else
                {
                    /* otherwise have small number */
                    p += "0";
                }

                // here we have i = number of 4digits printed
                // e = remains of exponent

                int n = App.P;
                while ((d_[n - 1] == 0) && n > i)
                    --n;

                if (i < n)
                {
                    if (point == 0)
                    {
                        point = p.Length;
                        p += ".";
                    }

                    for (; i < n; ++i)
                    {
                        v = d_[i];

                        p += v.ToString("D4");

                        pr += 4;

                        if ((precision != 0) && (pr >= precision))
                        {
                            int over = pr - precision;

                            p = p.Remove(p.Length - over);
                            break;
                        }
                    }
                }

                /* tidy up */
                if (point != 0)
                    while ((p.Length > point) && (p[p.Length - 1 - 1] == '0' || p[p.Length - 1 - 1] == '.'))
                        p = p.Remove(p.Length - 1);

                // print exponent
                if ((e != 0) || scimode)
                    p += string.Format("e{0}", e * 4 + eadj);
            }

            buf = p;
        }



        public int exp()
        {
            return GlobalMembersBcdfloat.GET_EXP(d_, App.P);
        }

        public void exp(int v)
        {
            GlobalMembersBcdfloat.SET_EXP(d_, App.P, v);
        }

        public bool neg()
        {
            return ((d_[App.P] & App.NEG) != 0) && (d_[0] != 0 || isInf());
        }

        public void setSign()
        {
            d_[App.P] |= App.NEG;
        }

        public void clearSign()
        {
            GlobalMembersBcdfloat.CLEAR_SIGN(d_, App.P);
        }

        public void negate()
        {
            d_[App.P] ^= App.NEG;
        }

        public bool isSpecial()
        {
            return GlobalMembersBcdfloat.GET_SPECIAL(d_, App.P);
        }

        public bool isZero()
        {
            return GlobalMembersBcdfloat.IS_ZERO(d_, App.P);
        }

        public bool isNan()
        {
            return GlobalMembersBcdfloat.GET_NAN(d_, App.P);
        }

        public bool isInf()
        {
            return GlobalMembersBcdfloat.GET_INF(d_, App.P);
        }

        public bool isInteger()
        {
            if (isZero())
                return true;

            int e = exp();
            int i;
            for (i = App.P - 1; i >= 0; --i)
                if (d_[i] != 0)
                    return e > i;
            return false;
        }

        public void ldexp(uint mant, int e)
        {
            // load the exp as a 4-block
            _init();
            _fromUInt(mant);
            exp(e);
        }



        public static void add(BCDFloat a, BCDFloat b, BCDFloat c)
        {
            GlobalMembersBcdfloat.bcd_add(a.d_, b.d_, c.d_, App.P);
        }

        public static void sub(BCDFloat a, BCDFloat b, BCDFloat c)
        {
            GlobalMembersBcdfloat.bcd_sub(a.d_, b.d_, c.d_, App.P);
        }

        public static void mul(BCDFloat a, BCDFloat b, BCDFloat c)
        {
            GlobalMembersBcdfloat.bcd_mul(a.d_, b.d_, c.d_, App.P);
        }

        public static void div(BCDFloat a, BCDFloat b, BCDFloat c)
        {
            GlobalMembersBcdfloat.bcd_div(a.d_, b.d_, c.d_, App.P);
        }



        public static bool sqrt(BCDFloat a, ref BCDFloat r)
        {
            if (a.neg() || a.isNan())
                return false;

            if (a.isInf())
            {
                r = a;  // sqrt(inf) = inf
                return true;
            }

            int rs;
            int v;
            bool rodd;
            int q;

            // acc, t and u get one more digit of workspace.
            ushort[] acc = new ushort[App.P + 2];
            int @as;

            ushort[] t = new ushort[App.P + 2];
            int ts;

            ushort[] u = new ushort[App.P + 2];
            int us = 0;

            r._init();

            int e = a.exp();

            v = (int) GlobalMembersBcdfloat.isqrt(a.d_[0]);

            rodd = (e & 1) == 0;
            r.d_[0] = (ushort)v;

            if (rodd)
                r.d_[0] *= 100;

            rs = 1;

            @as = 0;
            acc[0] = (ushort)(a.d_[0] - v * v);
            if (acc[0] != 0)
                ++@as;

            int i;
            int j = 1;
            int ca;

            for (; ; )
            {
                // bring in the next digit
                acc[@as] = (ushort)(j < App.P ? a.d_[j++] : 0);

                q = 0;
                if (acc[0] != 0)
                {
                    ++@as;

                    // t = 200*r if even, t=2*r if odd
                    int m = rodd ? 2 : 200;
                    ca = 0;
                    ts = rs;
                    for (i = rs; i > 0; --i)
                    {
                        v = ((int)r.d_[i - 1]) * m + ca;
                        ca = 0;
                        if (v >= App.BASE)
                        {
                            ca = v / App.BASE;
                            v -= ca * ((int)App.BASE);
                        }
                        t[i] = (ushort)v;
                    }
                    t[0] = (ushort)ca;
                    ++ts;

                    while (t[0] == 0)
                    {
                        for (i = 0; i < ts; ++i)
                            if (i < t.Length - 1)       // TO DO: Added array size check and shift in zero of over - verify
                                t[i] = t[i + 1];
                            else
                                t[i] = 0;
                        --ts;
                    }

                    if (ts > App.P + 1)
                    {
                        // rarely, the tempory size can become bigger than
                        // we can handle. this can only happen on the last
                        // digit. if so, stop.
                        //
                        break;
                    }

                    q = 0;
                    if (ts == @as)
                    {
                        if (ts > 1)
                            q = (((int)acc[0]) * ((int)App.BASE) + acc[1]) / (((int)t[0]) * ((int)App.BASE) + t[1]);
                        else
                            q = ((int)acc[0]) / t[0];
                    }
                    else if (@as > ts)
                    {
                        q = (((int)acc[0]) * ((int)App.BASE) + acc[1]) / t[0];
                    }

                    if (q != 0)
                    {
                        if (q > 99) q = 99;

                        /* t = t + q */
                        t[ts - 1] += (ushort)q;  // cant carry

                        for (; ; )
                        {
                            /* u = t*q */
                            ca = 0;
                            us = ts;
                            for (i = ts; i > 0; --i)
                            {
                                v = t[i - 1] * q + ca;
                                ca = 0;
                                if (v >= App.BASE)
                                {
                                    ca = v / App.BASE;
                                    v -= ca * ((int)App.BASE);
                                }
                                u[i] = (ushort)v;
                            }

                            u[0] = (ushort)ca;

                            if (ca != 0)
                                ++us;
                            else
                                for (i = 0; i < us; ++i)
                                    if (i < u.Length - 1)           // TO DO: added bounds check as safety, though haven't hit - verify
                                        u[i] = u[i + 1];
                                    else
                                        u[i] = 0;

                            // determine whether u > acc. if so then q was too
                            // big.
                            //
                            bool fail = us > @as;
                            if (!fail && us == @as)
                            {
                                for (i = 0; i < @as; ++i)
                                {
                                    int d = u[i] - acc[i];
                                    if (d > 0)
                                    {
                                        fail = true;
                                        break;
                                    }
                                    else if (d < 0)
                                        break;
                                }
                            }

                            if (!fail) break;

                            // decrease q by 1 and try again
                            q -= 1;
                            --t[ts - 1]; // adjust for new q
                        }
                    }
                }

                if (rodd)
                {
                    // can accommodate 2 more digits in current size
                    r.d_[rs - 1] += (ushort)q;
                    rodd = false;

                    if (rs == App.P + 1)
                        break;
                }
                else
                {
                    r.d_[rs++] = (ushort)(q * 100);
                    rodd = true;
                }

                if (q != 0)
                {
                    // acc = acc - u.
                    // wont borrow because u <= acc.
                    //
                    int k;
                    ca = 0;
                    k = us;
                    for (i = @as - 1; i >= 0; --i)
                    {
                        v = acc[i] - ca;
                        if (k > 0) v -= u[--k];
                        ca = 0;
                        if (v < 0)
                        {
                            v += App.BASE;
                            ca = 1;
                        }
                        acc[i] = (ushort)v;
                    }

                    while (acc[0] == 0)
                    {
                        for (i = 0; i < @as; ++i)
                            if (i < acc.Length - 1)         // TO DO: added bounds check
                                acc[i] = acc[i + 1];
                            else
                                acc[i] = 0;

                        if (--@as == 0)
                            break;
                    }
                }
            }

            e = e >= -1 ? (e + 1) / 2 : e / 2;

            if (r._round25() != 0)
                ++e;

            r.exp(e);
            return true;
        }



        public static bool lt(BCDFloat a, BCDFloat b)
        {
            /* true iff a < b */
            return GlobalMembersBcdfloat.bcd_cmp(a.d_, b.d_, App.P) < 0;
        }

        public static bool le(BCDFloat a, BCDFloat b)
        {
            /* true iff a <= b */
            return GlobalMembersBcdfloat.bcd_cmp(a.d_, b.d_, App.P) <= 0;
        }

        public static bool gt(BCDFloat a, BCDFloat b)
        {
            /* true iff a > b */
            return GlobalMembersBcdfloat.bcd_cmp(a.d_, b.d_, App.P) == 1;
        }

        public static bool ge(BCDFloat a, BCDFloat b)
        {
            /* true iff a >= b */
            int res = GlobalMembersBcdfloat.bcd_cmp(a.d_, b.d_, App.P);
            return res == 0 || res == 1;
        }

        public static bool equal(BCDFloat a, BCDFloat b)
        {
            return GlobalMembersBcdfloat.bcd_cmp(a.d_, b.d_, App.P) == 0;
        }

        public static int ifloor(BCDFloat x)
        {
            BCDFloat a = new BCDFloat();
            floor(x, ref a);
            return a.asInt();
        }

        public static int itrunc(BCDFloat x)
        {
            BCDFloat a = new BCDFloat();
            trunc(x, ref a);
            return a.asInt();
        }

        public static bool floor(BCDFloat a, ref BCDFloat c)
        {
            /* floor, largest integer <= a.
             * eg floor(2.1) = 2.
             *    floor(-2.1) = -3.
             */

            c = new BCDFloat(a);
            int e = c.exp();
            int i;
            bool changed = false;
            for (i = App.P - 1; i >= 0; --i)
            {
                if (i >= e)
                {
                    if (c.d_[i] != 0)
                        changed = true;

                    c.d_[i] = 0;
                }
            }

            if (c.neg() && changed)
            {
                /* need to subtract 1 */
                for (i = App.P - 1; i >= 0; --i)
                {
                    if (c.d_[i] != 0)
                    {
                        ++c.d_[i];
                        break;
                    }
                }
            }
            return true;
        }

        public static bool trunc(BCDFloat a, ref BCDFloat c)
        {
            /* truncate towards zero.
             * trunc(2.1) = 2.
             * trunc(-2.1) = -2
             */
            c = new BCDFloat(a);
            int e = c.exp();
            int i;
            for (i = App.P - 1; i >= 0; --i)
                if (i >= e)
                    c.d_[i] = 0;

            return true;
        }



        public void _init()
        {
            for (int i = 0; i <= App.P; ++i)
                d_[i] = 0;
        }



        public static void _uadd(BCDFloat a, BCDFloat b, BCDFloat c)
        {
            GlobalMembersBcdfloat.bcd_uadd(a.d_, b.d_, c.d_, App.P);
        }



        public int _round25()
        {
            return GlobalMembersBcdfloat.bcd_round25(d_, App.P);
        }

        public void _fromUInt(uint v)
        {
            GlobalMembersBcdfloat.bcd_fromUInt(d_, App.P, v);
        }

        public void _roundDigits(uint precision, ref BCDFloat v)
        {
            if (!isSpecial() && !isZero() && precision < 25)
            {
                bool n = neg();

                BCDFloat rv = new BCDFloat();

                // adjust precision to number of digits before 4radix.
                if (d_[0] >= 10) --precision;
                if (d_[0] >= 100) --precision;
                if (d_[0] >= 1000) --precision;
                --precision; // units

                int m = GlobalMembersBcdfloat.BCDDecade[precision & 3] * 5;
                rv.ldexp((uint)m, (int)(exp() - (precision >> 2) - 1));

                // ignores signs
                _uadd(this, rv, v);

                // restore sign if needed
                if (n)
                    v.setSign();
            }
            else
                v = this;
        }



        public static BCDFloat posInf()
        {
            return new BCDFloat(posInfD_);
        }

        public static BCDFloat negInf()
        {
            return new BCDFloat(negInfD_);
        }

        public static BCDFloat nan()
        {
            return new BCDFloat(nanD_);
        }



        public static void epsilon(int n, BCDFloat v)
        {
            int m = GlobalMembersBcdfloat.BCDDecade[(n - 1) & 3];
            v.ldexp((uint)m, -(n >> 2));
        }

        public static void mul2(ushort[] ad, int ea, ushort[] bd, int eb, ushort[] cd, ref int ec)
        {
            int ca;
            int i, j;
            int u, v;

            ushort[] acc = new ushort[2 * App.P+1];

            for (i = 0; i < 2 * App.P; ++i)
                cd[i] = 0;

            int cc = 0;
            for (i = 2 * App.P - 1; i >= 0; --i) 
            {
                for (j = 2 * App.P; j > 0; --j)
                    cd[j] = cd[j-1];

                cd[0] = (ushort) cc;
                cc = 0;
                
                u = ad[i];
                if (u == 0)
                    continue;

                ca = 0;
                for (j = 2 * App.P; j > 0; --j) 
                {
                    v = bd[j-1]*u + ca;
                    ca = 0;
                    if (v >= App.BASE) 
                    {
                        ca = v / App.BASE;
                        v = v - ca * ((int)App.BASE);
                    }
                    acc[j] = (ushort) v;
                }
                acc[0] = (ushort) ca;

                /* now add acc into c */
                for (j = 2 * App.P; j >= 0; --j) 
                {
                    v = cd[j] + acc[j] + cc;
                    cc = 0;
                    if (v >= App.BASE) 
                    {
                        cc = 1;
                        v -= App.BASE;
                    }
                    cd[j] = (ushort) v;
                }
            }
            cd[0] += (ushort) cc;  // carry?
            ec = ea + eb;
        }

        public int asInt()
        {
            // if we fit in an int, return it otherwise 0
            int ea = exp();
            int v = 0;
            int i = 0;
            while (i < ea && i < App.P)
            {
                if (v > 214748L) return 0; // too large, bail out.
                v *= App.BASE;
                v += d_[i];
                ++i;
            }
            if (neg()) v = -v;
            return v;
        }

        #endregion
    }
}
